#  ATAudio: provides handling of audio files in a Plone site
#  Copyright (c) 2003-2006 Nate Aune
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""ATAudio allows you to manage audio files in your Plone site.
   Stores the files on the file system, instead of in the ZODB.
   Extracts the ID3 tags from the MP3 file as metadata.
"""
__author__  = 'Nate Aune <natea@jazkarta.com>'
__docformat__ = 'restructuredtext'
from Products.CMFCore import CMFCorePermissions
from Products.CMFCore.utils import getToolByName

from AccessControl import ClassSecurityInfo
from ZODB.PersistentMapping import PersistentMapping
from Acquisition import aq_base

from config import *
import exceptions


from Products.Archetypes.Storage import Storage

try:
    from Products.LinguaPlone.public import *
except:
    # no multilingual support
    from Products.Archetypes.public import *

if HAS_EXT_STORAGE and EXT_STORAGE_ENABLE:
    from Products.ExternalStorage.ExternalStorage import ExternalStorage
else:
    raise exceptions.DependancyError('Missing product: ExternalStorage')
 
# ATContentTypes 1.0
if HAS_ATCT_1_0:
    from Products.ATContentTypes.content.file import ATFile
    from Products.ATContentTypes.content.file import ATFileSchema
    from Products.ATContentTypes.content.base import registerATCT
else:
# ATContentTypes 0.2 - maintain backwards compatibility for ATCT 0.2
    from Products.ATContentTypes.types.ATFile import ATFile
    from Products.ATContentTypes.types.schemata import ATFileSchema

from cStringIO import StringIO

import sys
import os
from Globals import package_home
 
# eyeD3 imports to assist in the handling of ID3 tags
sys.path.append(os.path.join(package_home(GLOBALS), 'utils'))
import eyeD3
from eyeD3.tag import genres, TagException
from eyeD3.binfuncs import *
from eyeD3.utils import *
from eyeD3.frames import DEFAULT_ENCODING, UTF_8_ENCODING, UTF_16_ENCODING, \
     IMAGE_FID, id3EncodingToString, FrameHeader, ImageFrame


# extend genres with our custom genres in config.py
# just in case we refresh the product
# to prevent multipling custom genres each time we refresh
# check the length of genres

OLD_EYED3_GENRE_MIN = genres.GENRE_MAX
if genres.EYED3_GENRE_MIN - OLD_EYED3_GENRE_MIN < len(ADDITIONAL_GENRES):
    i = 1
    for key in ADDITIONAL_GENRES:
        genres[genres.EYED3_GENRE_MIN+i] = key
        i += 1
    genres.EYED3_GENRE_MIN = genres.EYED3_GENRE_MIN + len(ADDITIONAL_GENRES)


# If we have Plodcasting we can use Plodcasting specific fields.

if HAS_PLODCASTING:
    plodcastingFields=()
else:
    plodcastingFields=(
        BooleanField('cacheStatus',
                     required=1,
                     default=True,
                     widget=BooleanWidget(label="Allow Caching",
                                          description="Allow external caching of this object",
                                          label_msgid="label_audiofile_cacheStatus",
                                          i18n_domain="ataudio"),
                     schemata="Metadata",
                     ),
    )

ATAudioSchema = ATFileSchema.copy() + Schema(plodcastingFields).copy() + Schema((

    FileField('file',
                primary=1,
                required=1,
                languageIndependent=1,
                validators=MaxSizeValidator("checkFileMaxSize", maxsize=MAX_FILE_SIZE),
                storage=ExternalStorage(prefix=FS_DIRECTORY, archive=False, rename=False),
                widget=FileWidget(label="Audio File",
                                    label_msgid="label_audiofile_file",
                                    description="Upload your MP3 file and if it has ID3 tags, the fields on this form will be auto-populated.",
                                    description_msgid="desc_audiofile_file",
                                    i18n_domain="ataudio")
                ),

    StringField('title',
                searchable=1,
                required=False,
                languageIndependent=1,
                default_method="getAudioTitle",
                accessor="Title",
                widget=StringWidget(label="Title",
                                    label_msgid="label_audiofile_title",
                                    description="The title of this audio file.",
                                    description_msgid="desc_audiofile_title",
                                    i18n_domain="ataudio"),
                schemata="Metadata",
                ),

    StringField('audioTitle',
                searchable=1,
                languageIndependent=1,
                widget=StringWidget(label="Audio file title",
                                    label_msgid="label_audiofile_audiotitle",
                                    description="The title of this audio file.",
                                    description_msgid="desc_audiofile_title",
                                    i18n_domain="ataudio"),
                schemata="Metadata",
                ),

    StringField('artist',
                searchable=1,
                languageIndependent=1,
                widget=StringWidget(label="Artist",
                                    label_msgid="label_audiofile_artist",
                                    description="The artist of this audio file.",
                                    description_msgid="desc_audiofile_artist",
                                    i18n_domain="ataudio"),
                schemata="Metadata",
                ),

    StringField('album',
                searchable=1,
                languageIndependent=1,
                widget=StringWidget(label="Album",
                                    label_msgid="label_audiofile_album",
                                    description="The album that this audio file is on",
                                    description_msgid="desc_audiofile_album",
                                    i18n_domain="ataudio"),
                schemata="Metadata",
                ),

    IntegerField('year',
                searchable=1,
                languageIndependent=1,
                validators=('isInt',),
                widget=IntegerWidget(label="Year",
                                     label_msgid="label_audiofile_year",
                                     description="Enter the year that this audio file was recorded.",
                                     description_msgid="desc_audiofile_year",
                                     i18n_domain="ataudio",
                                     size="4",
                                     ),
                schemata="Metadata",
                ),

    StringField('genre',
                searchable=1,
                languageIndependent=1,
                widget=SelectionWidget(label="Genre",
                                     label_msgid="label_audiofile_genre",
                                     description="Enter the genre that this audio file was recorded.",
                                     description_msgid="desc_audiofile_genre",
                                     i18n_domain="ataudio",
                                     ),
                vocabulary='getGenres',
                schemata="Metadata",
                ),

    TextField('comment',
                searchable=1,
                languageIndependent=1,
                edit_accessor='getComment',
                widget=TextAreaWidget(label="Comment",
                                     label_msgid="label_audiofile_comment",
                                     description="Enter a comment about this audio file.",
                                     description_msgid="desc_audiofile_comment",
                                     i18n_domain="ataudio",
                                     ),
                schemata="Metadata",
                ),

    ImageField('album_art',
                mutator='setImage',
                languageIndependent=1,
                widget=FileWidget(label="Album Art",
                                     label_msgid="label_album_art",
                                     description_msgid="desc_album_art",
                                     i18n_domain="ataudio",
                                     ),
                schemata="Metadata",
                ),
    ),
    marshall=PrimaryFieldMarshaller(),
   )

new_actions = (
    {'id': 'play',
     'name': 'Play',
     'action': 'string:${object_url}/stream',
     'permissions': (CMFCorePermissions.View,),
     'category': 'document_actions',
     },
    {'id': 'download',
     'name': 'Download',
     'action': 'string:${object_url}/download',
     'permissions': (CMFCorePermissions.View,),
     'category': 'document_actions',
     },
    )

def modify_fti(fti):
    fti['actions'] = fti['actions'] + new_actions

def getAudioFilePath(self):
    """return the file path"""
    path = self.getField('file').storage.getFilepath(self,'file')

    if not os.path.isfile(path):
        log("ATAudio: error, trying to get tag of non-file: %s" % path)
        raise AttributeError("MP3 file not created, no ID3 tag available")
    else:
        return path

def getAudioFileSize(self):
    # Is this a sensible default?
    result = 0

    path = self.getField('file').storage.getFilepath(self,'file')
    if path:
    # is there a more Pythonic way to get the size of the file? NJA
        s = os.stat(path)
        result = s[6]

    return result

def hasAudioData(self):
    result = None

    size = getAudioFileSize(self)
    result = size > 0

    return result


class ATAudio(ATFile):
    """
    A content type that handles audio files in your Plone site.
    """

    schema         =  ATAudioSchema

    content_icon   = 'audio_icon.gif'
    portal_type    = 'ATAudio'
    meta_type      = 'ATAudio'
    archetype_name = 'Audio file'
    immediate_view = 'audio_view'
    default_view   = 'audio_view'
    suppl_views    = ()
    _atct_newTypeFor = None
    assocMimetypes = ('audio/*', )
    assocFileExt   = ()

    security       = ClassSecurityInfo()

    _at_rename_after_creation = True

    actions = (
       {'id'            : 'view',
        'name'          : 'View',
        'action'        : 'string:${object_url}/audio_view',
        'permissions'   : (CMFCorePermissions.View,)
        },
       )

    def __init__(self, *args, **kwargs):
        self.schema['allowDiscussion'].default=1
        self.schema['id'].widget.visible = {'edit': 'invisible'}
        ATFile.__init__(self, *args, **kwargs)
        self.mp3header = None
        self.id3tags = None


    security.declareProtected(CMFCorePermissions.View, 'index_html')
    def index_html(self):
        """Allows file direct access download."""
        field = self.getPrimaryField()
        field.download(self)

    def getAudioURL(self, media_server=None):
        """
        get the URL for the audio file. Optionally we can pass in the url for a media server which is assumed to be
        a regular web-server with a similar directory structure to the zope instance.
        """
        # relative_url='/'.join(getToolByName(self, 'portal_url').getRelativeContentPath(self))
        ptool = getToolByName(self, 'portal_url')
        portal_url=ptool()

        relative_url = ptool.getRelativeContentURL(self)

        out = StringIO()

        if media_server:
            # This allows us to host Audio files via some other mechanism beside Zope.
            out.write('http://%s/%s' % (media_server, relative_url))
        else:
            # Just a normal URL, we assume that the file is served from Zope.
            out.write('%s/%s' % (portal_url, relative_url))

        return out.getvalue()

    def stream(self, REQUEST, RESPONSE):
        """stream the audio file as M3U"""
        RESPONSE.setHeader('Content-Type', self.getMimeType() )
        RESPONSE.setHeader('Content-Disposition', 'attachment; filename="%s.m3u"' % self.getId()[:-4])
        out = StringIO()
        out.write(REQUEST.URL1 + '\n')
        return out.getvalue()

    def getTags(self):
        """ get the Tag linked """
        if not getattr(self, 'id3tags', None):
            self._setTags()
        if self.id3tags and not self.id3tags.frames:
            self._setTags()
            if not self.id3tags.frames:
                self.id3tags.setVersion(ID3_DEFAULT_VERSION)
                #return None
        return self.id3tags

    def _setTags(self):
        """ get the Tag linked """
        self.id3tags = eyeD3.Tag()
        try:
            path = getAudioFilePath(self)
            link = self.id3tags.link(path)
        except:
            pass
        self._p_changed = 1
            #self.id3tags = None
 
    def getMP3Header(self):
        """ get the Mp3AudioFile instance for the file"""
        if not getattr(self, 'mp3header', None):
            self._setMP3Header()
        return self.mp3header

    def _setMP3Header(self):
        """ """
        try:
            path = getAudioFilePath(self)
            self.mp3header = eyeD3.Mp3AudioFile(path)
        except:
            self.mp3header = None
        self._p_changed = 1

    """ temporary needed for eyeD3 migration 0.6.6 -> 0.6.7
        call it to reset cached attributes values
        smooth update/migration should be done on reinstall """
    def forceTagsReset(self):
        """ reset mp3header & id3tags"""
        self._setMP3Header()
        self._setTags()

    def getGenres(self):
        """ list genres for popup """
        values = [str(i) for i in range(genres.EYED3_GENRE_MIN+1)]
        return DisplayList(zip(values, genres))

    def Title(self):
        """get the title from the audio title"""
        tags = self.getTags()
        if tags:
            value = tags.getTitle()
            if isinstance(value, unicode):
                value = value.encode(self.getCharset())
            return value
        return ''

    def getAlbum(self):
        """get the album"""
        tags = self.getTags()
        if tags:
            value = tags.getAlbum()
            return value
        return ''

    def getArtist(self):
        """get the artist"""
        tags = self.getTags()
        if tags:
            value = tags.getArtist()
            return value
        return ''

    def getYear(self):
        """get the year"""
        tags = self.getTags()
        if tags:
            value = tags.getYear()
            return value
        return ''

    def getAudioTitle(self):
        """get the title"""
        tags = self.getTags()
        if tags:
            value = tags.getTitle()
            return value
        return ''

    def getGenre(self):
        """get the genre"""
        tags = self.getTags()
        value = ''
        if tags:
            genre = tags.getGenre()
            if genre and genre.getId() is not None:
                value = int(genre.getId())
        return value

    def getComment(self):
        """get the comment"""
        tags = self.getTags()
        if tags:
            value = tags.getComments()
            if len(value)>0:
                return value[0].comment
        return ''


    """
    Stream information, should not be directly mutable.
    """
    def getBitRate(self):
        """
        Get the bitrate.
        This is a stub method, for the sake of Plodcasting.
        """
        mp3header = self.getMP3Header()
        if mp3header:
            # the result can also be formated from mp3header.getBitRate()
            # it is 2 elements tuple  (isVariableBitrate, Bitrate)
            return mp3header.getBitRateString()
        return None

    def getFrequency(self):
        """
        Get the audio sampling frequency of the object. Currently not implemented.
        """
        mp3header = self.getMP3Header()
        if mp3header:
            return mp3header.getSampleFreq()
        return None

    def getLength(self):
        """
        Get the length of the audio/video stream in seconds.
        """
        mp3header = self.getMP3Header()
        if mp3header:
            return mp3header.getPlayTime()
        return None

    def getLengthString(self):
        """
        Get the length of the audio/video stream formated hour:minute:second.
        """
        mp3header = self.getMP3Header()
        if mp3header:
            return mp3header.getPlayTimeString()
        return None

    def getFileSize(self):
        """
        Get the total size of the audio file in bytes.
        This is required for Plodcasting
        """
        return getAudioFileSize(self)

    def getMimeType(self):
        """
        Return the mime type of the audio/video file as a string
        This is required by Plodcasting to produce compliant RSS feeds.
        """
        return self.Schema()['file'].getBaseUnit(self).mimetype

    def strToUnicode(self, s):
        """
        Convert a byte string in Plone's site encoding into a Unicode string.
        """
        if isinstance(s, str):
            return s.decode(self.getCharset())
        else:
            return s
        
    """
    Mutators.
    """
    def setArtist(self, v, **kwargs):
        """Set the artist"""
        tags = self.getTags()
        if tags:
            tags.setArtist(self.strToUnicode(v))

    def setTitle(self, v, **kwargs):
        """Set the title"""
        tags = self.getTags()
        if tags:
            tags.setTitle(self.strToUnicode(v))

    def setAlbum(self, v, **kwargs):
        """Set the album"""
        tags = self.getTags()
        if tags:
            tags.setAlbum(self.strToUnicode(v))


    def setYear(self, v, **kwargs):
        """Set the year"""
        tags = self.getTags()
        if tags:
            tags.setDate(v)


    def setComment(self, v, **kwargs):
        """Set the comment"""
        tags = self.getTags()
        if tags:
            comments = tags.frames["COMM"]
            for c in comments:
                tags.frames.remove(c)
            tags.addComment(self.strToUnicode(v))


    def setGenre(self, v, **kwargs):
        """Set the genre"""
        tags = self.getTags()
        if tags:
            tags.setGenre(v)


    # the file seems to be set properly in this way but the
    # instance id is not updated
    #
    #def setFile(self, v, **kw):
    #    """Set the file field
    #       Setup tags & headers
    #    """
    #    self.getField('file').set(self, v, **kw)
    #    self.id3tags = self._setTags()
    #    self.mp3header = self._setMP3Header()

    def update(self, **kwargs):
        """ overiding the update of BaseObject
            to update all tags in one file write
        """
        self.Schema().updateAll(self, **kwargs)
        self._p_changed = 1
        tags = self.getTags()
        if tags:
            tags.update()
        self.reindexObject()

    def getImage(self, REQUEST=None, RESPONSE=None):
        """ return the Album Art first image """
        tags = self.getTags()
        if not tags:
            return None
        imgs = tags.getImages()
        img = None
        tags.clear()
        if len(imgs)>0:
            img = imgs[0]
        if img:
            if not RESPONSE:
                RESPONSE = self.REQUEST.RESPONSE
            RESPONSE.setHeader('Content-Type', img.mimeType)
            return img.imageData
        return None

    def hasImage(self):
        """ check if Album Art exists """
        tags = self.getTags()
        if not tags:
            return None
        imgs = tags.getImages()
        img = None
        if len(imgs)>0:
            img = imgs[0]
            if img:
                return 1
        return None

    def setImage(self, value=None, desc = u"", encoding=DEFAULT_ENCODING, **kwargs):
        """ set the Album Art first image FRONT_COVER """
        if value:
            tags = self.getTags()
            if not tags:
                return None
            imgData = value.read()
            mt = value.headers['content-type']
            frameData = DEFAULT_ENCODING
            frameData += mt + "\x00"
            frameData += bin2bytes(dec2bin(3, 8))
            frameData += desc.encode(id3EncodingToString(encoding)) + "\x00"
            frameData += imgData
            frameHeader = FrameHeader()
            frameHeader.id = IMAGE_FID
            frame = ImageFrame(frameHeader, data = frameData);
            imgs = tags.getImages()
            if len(imgs) == 0:
                tags.frames.append(frame)
            else:
                for i in tags.frames:
                    if i == imgs[0]:
                        index = tags.frames.index(i)
                        tags.frames[index] = frame
            tags.update()
            tags.clear()

    def migrate(self):
        """Migrates data from the old format to the new one.
        """
        storage = self.getField('file').storage
        filename = self.getId()
        try:
            filepath = storage.getFileSystemPath(self, filename)
            f = open(filepath,'rb')
            self.setFile(f)
            f.close()
            self.forceTagsReset()
            return "Success %s" % self.getId()
        except:
            return "Failed %s" % self.getId()

    def processForm(self, data=1, metadata=0, REQUEST=None, values=None):
        """Process the schema looking for data in the form"""
        self._processForm(data=data, metadata=metadata,
                          REQUEST=REQUEST, values=values)
        tags = self.getTags()
        if tags:
            if tags.getVersion() >= ID3_V2_4:
                tags.setTextEncoding(UTF_8_ENCODING)
            elif tags.getVersion() >= ID3_V2:
                tags.setTextEncoding(UTF_16_ENCODING)

            tags.update()
            tags.clear()


registerType(ATAudio, PROJECTNAME)

